<?php
/**
 * PHP version 5
 *
 * @package    UASparser
 * @author     Jaroslav Mallat (http://mallat.cz/)
 * @copyright  Copyright (c) 2008 Jaroslav Mallat
 * @copyright  Copyright (c) 2010 Alex Stanev (http://stanev.org)
 * @copyright  Copyright (c) 2012 Martin van Wingerden (http://www.copernica.com)
 * @copyright  Copyright (c) 2012 Fernando Zorrilla de San Martin (http://wpgetready.com)
 * @version    0.6
 * @license    http://www.gnu.org/licenses/lgpl.html GNU Lesser General Public License
 * @link       http://user-agent-string.info/download/UASparser
 *
 *
 * 20120404: Considerations
 * - cache folder needs write configuration to work properly.
 * - Return info with following format:
 * Array
(
    [typ] => browser
    [ua_family] => Firefox
    [ua_name] => Firefox 3.0.8
    [ua_url] => http://www.mozilla.org/products/firefox/
    [ua_company] => Mozilla Foundation
    [ua_company_url] => http://www.mozilla.org/
    [ua_icon] => firefox.png
    [ua_info_url] => http://user-agent-string.info/list-of-ua/browser-detail?browser=Firefox
    [os_family] => Windows
    [os_name] => Windows XP
    [os_url] => http://www.microsoft.com/windowsxp/default.mspx
    [os_company] => Microsoft Corporation.
    [os_company_url] => http://www.microsoft.com/
    [os_icon] => windowsxp.png
)
 * 20120410: v1.6.31; disabled _downloaddata because the solution does not scale:
 * 			 a new solution is in development, to be launched in v1.6.4
 * 20120418: v1.6.5 Complete rewrite of _loadData and _downloadData
 *                 Replaced by _masterDownload
 *                 Advantages:
 *                      -Much improved control over download process.
 *                      -Use of HTTP API
 *                      -Documented process
 *                      -Using a new service providing more reliability than user-agent-string.info
 *                      and avoiding blacklisting IP.
 *                      -Zip process file, means even faster download.
 * To do: - Control Panel
 *        - Ajax control
 *        - Options control.
 **/

 // view source this file and exit
//if ($_GET['source'] == "y") {     show_source(__FILE__);     exit; }


abstract class UASparser
{
    public $updateInterval =   1209600; // 2 weeks

    private static $_zip_url    =   'http://statcomm.wpgetready.com/statservice/getfile';
    private static $_ver_url    =   'http://statcomm.wpgetready.com/statservice/version';
    private static $_md5_url    =   'http://statcomm.wpgetready.com/statservice/md5';
    private static $_info_url   =   'http://user-agent-string.info';
    private static $TIMEOUT     = 3; //general timeout for operations.
    private static $plugin_version =   "v1.6.70";

    private $_cache_dir         =   null;
    private $_data              =   null;

    /**
     *  Constructor with an optional cache directory
     * @param null $cacheDirectory base directory
     * @param null $updateInterval update interval in seconds to refresh user agent data.
     * @internal param \cache $string directory to be used by this instance
     */
    public function __construct($cacheDirectory = null, $updateInterval = null) {
        if ($cacheDirectory) $this->SetCacheDir($cacheDirectory);
        $options=get_option("statcomm_options");

        //if ($updateInterval) $this->updateInterval =  $options['cmb_uarefresh'];
        $this->updateInterval =  isset($options['current_lapse'])?$options['current_lapse']:'1209600';
    }

    /**
     *  Parse the useragent string if given otherwise parse the current user agent
     *  @param  string  useragent string
     * @return array
     */
    public function Parse($useragent = null) {
       // fl("backtrace:", var_dump(debug_backtrace()));
        // intialize some variables
        $browser_id = $os_id = null;
        $result = array();

        // initialize the return value
        $result['typ']              = 'unknown';
        $result['ua_family']        = 'unknown';
        $result['ua_name']          = 'unknown';
        $result['ua_version']       = 'unknown';
        $result['ua_url']           = 'unknown';
        $result['ua_company']       = 'unknown';
        $result['ua_company_url']   = 'unknown';
        $result['ua_icon']          = 'unknown.png';
        $result["ua_info_url"]      = 'unknown';
        $result["os_family"]        = 'unknown';
        $result["os_name"]          = 'unknown';
        $result["os_url"]           = 'unknown';
        $result["os_company"]       = 'unknown';
        $result["os_company_url"]   = 'unknown';
        $result["os_icon"]          = 'unknown.png';

        // if no user agent is supplied process the one from the server vars
        if (!isset($useragent) && isset($_SERVER['HTTP_USER_AGENT'])){
            $useragent = $_SERVER['HTTP_USER_AGENT'];
        }

        // if we haven't loaded the data yet, do it now

        if(!$this->_data) {
            $this->_data = $this->_loadData();
        }

        // we have no data or no valid user agent, just return the default data
        if(!$this->_data || !isset($useragent)) {
            return $result;
        }

        // crawler
        foreach ($this->_data['robots'] as $test) {
            if ($test[0] == $useragent) {
                $result['typ']                            = 'Robot';
                if ($test[1]) $result['ua_family']        = $test[1];
                if ($test[2]) $result['ua_name']          = $test[2];
                if ($test[3]) $result['ua_url']           = $test[3];
                if ($test[4]) $result['ua_company']       = $test[4];
                if ($test[5]) $result['ua_company_url']   = $test[5];
                if ($test[6]) $result['ua_icon']          = $test[6];
                if ($test[7]) { // OS set
                    $os_data = $this->_data['os'][$test[7]];
                    if ($os_data[0]) $result['os_family']       =   $os_data[0];
                    if ($os_data[1]) $result['os_name']         =   $os_data[1];
                    if ($os_data[2]) $result['os_url']          =   $os_data[2];
                    if ($os_data[3]) $result['os_company']      =   $os_data[3];
                    if ($os_data[4]) $result['os_company_url']  =   $os_data[4];
                    if ($os_data[5]) $result['os_icon']         =   $os_data[5];
                }
                if ($test[8]) $result['ua_info_url']      = self::$_info_url.$test[8];
                return $result;
            }
        }

        // find a browser based on the regex
        foreach ($this->_data['browser_reg'] as $test) {
            if (@preg_match($test[0],$useragent,$info)) { // $info may contain version
                $browser_id = $test[1];
                break;
            }
        }

        // a valid browser was found
        if ($browser_id) { // browser detail
            $browser_data = $this->_data['browser'][$browser_id];
            if ($this->_data['browser_type'][$browser_data[0]][0]) $result['typ']    = $this->_data['browser_type'][$browser_data[0]][0];
            if (isset($info[1]))    $result['ua_version']     = $info[1];
            if ($browser_data[1])   $result['ua_family']      = $browser_data[1];
            if ($browser_data[1])   $result['ua_name']        = $browser_data[1].(isset($info[1]) ? ' '.$info[1] : '');
            if ($browser_data[2])   $result['ua_url']         = $browser_data[2];
            if ($browser_data[3])   $result['ua_company']     = $browser_data[3];
            if ($browser_data[4])   $result['ua_company_url'] = $browser_data[4];
            if ($browser_data[5])   $result['ua_icon']        = $browser_data[5];
            if ($browser_data[6])   $result['ua_info_url']    = self::$_info_url.$browser_data[6];
        }

        // browser OS, does this browser match contain a reference to an os?
        if (isset($this->_data['browser_os'][$browser_id])) { // os detail
            $os_id = $this->_data['browser_os'][$browser_id][0]; // Get the os id
            $os_data = $this->_data['os'][$os_id];
            if ($os_data[0])    $result['os_family']      = $os_data[0];
            if ($os_data[1])    $result['os_name']        = $os_data[1];
            if ($os_data[2])    $result['os_url']         = $os_data[2];
            if ($os_data[3])    $result['os_company']     = $os_data[3];
            if ($os_data[4])    $result['os_company_url'] = $os_data[4];
            if ($os_data[5])    $result['os_icon']        = $os_data[5];
            return $result;
        }

        // search for the os
        foreach ($this->_data['os_reg'] as $test) {
            if (@preg_match($test[0],$useragent)) {
                $os_id = $test[1];
                break;
            }
        }

        // a valid os was found
        if ($os_id) { // os detail
            $os_data = $this->_data['os'][$os_id];
            if ($os_data[0]) $result['os_family']       = $os_data[0];
            if ($os_data[1]) $result['os_name']         = $os_data[1];
            if ($os_data[2]) $result['os_url']          = $os_data[2];
            if ($os_data[3]) $result['os_company']      = $os_data[3];
            if ($os_data[4]) $result['os_company_url']  = $os_data[4];
            if ($os_data[5]) $result['os_icon']         = $os_data[5];
        }
        return $result;
    }


    /**
     * Load the file, if we need to donwload data delegate the task to _masterDownload
     */
    private function _loadData(){
        $PARSER_FILE  = $this->_cache_dir .'/uasdata.ini';
        //Handle all the process
        //Find if the service is enabled, if not skip download process
        $options=get_option("statcomm_options");
        $chk =isset($options['chk_uas'])?$options['chk_uas']:'';
        if($chk) {
            utilities::fl("UAS Service enabled, proceeding to test download");
            $this->_masterDownload();
        }
        else
        {
            utilities::fl("UAS service disabled, skipped");
        }

        if (file_exists($PARSER_FILE))
        {
            return @parse_ini_file($PARSER_FILE,true);
        }
        else
        {
            utilities::fl("there was an error parsing the uasdata.ini file");
        }
    }

    /**
     * handles the entire download procedure and decides when to download.
     * @return mixed
     */
	private function _masterDownload()
	{
		$PARSER_FILE  = $this->_cache_dir .'/uasdata.ini';
		$PARSER_CACHE = $this->_cache_dir.'/cache.ini';
		$PARSER_ZIP   = $this->_cache_dir.'/uasdata.zip';

        utilities::fl("checking uaparser existance...","");
		//Does the uaparser exists?
		if (file_exists($PARSER_FILE)) {
            utilities::fl("file $PARSER_FILE ok,checking cache","");
			//And also, does the cache exists?
			if (file_exists($PARSER_CACHE)) {
                utilities::fl("file $PARSER_CACHE ok, checking if time to update","");
				//Is it time to update?
				$cacheIni = parse_ini_file($PARSER_CACHE);
				//check last update. If does not exist, start download process (that's why we assign 0 if lastupdate undefined)
				$lastupdate = isset($cacheIni['lastupdate'])?$cacheIni['lastupdate']:0;
				//Additional check: $lastupdate has to be numeric
				//If difference between now and last update is less than interval, we've done here
                utilities::fl("seconds remaining before checking:", $this->updateInterval - time() + $lastupdate);
				if (time()- $lastupdate < $this->updateInterval)
				{	
					//Log: File and cache up to date, nothing to do.
                    utilities::fl("File up to date,nothing to do","");
					return;
				}
                //It is time to check the server.
                //'WordPress/' . $wp_version . '; ' . get_bloginfo( 'url' )
                $version_server= self::remoteRequest(self::$_ver_url);
                if ($this->versionError($version_server))
                    return;

                $ver=wp_remote_retrieve_body($version_server);
                $cache_ver=isset($cacheIni['localversion'])?$cacheIni['localversion']:"";
                utilities::fl("remote version:$ver local version: $cache_ver");
                //If we found that is the same version, then we don't have nothing to download.
                //TODO:We need to check that the server version is newer (not any version)
                if ($ver==$cache_ver)
                {
                    utilities::fl("found same version, rebuilding cache","");
                    //Rebuild cache file. Note that also reset time, so we don't have to query the
                    //service for predefined period
                    $this->buildCacheFile($PARSER_CACHE,$cache_ver);
                    return;
                }
                utilities::fl("Different version, rebuilding user agent data and cache","");

			}
            else
            {
                utilities::fl("cache not found, proceeding to download data","");
            }
		}
		//We come here under three conditions:
		//1-Parser file does not exists. It happens only if we delete the file manually, or in future versions if the
        //plugin comes without uasdata.ini
		//2-Parse file exists but cache does not. Possible explanations: uaparser.ini exists and folder restrictions
        //does not allow tu create cache. Very rare condition but it could happen. Also would happen if
        //we manually remove cache.ini.
		//3-Both file exists but it is time to update.

        //Before doing anything, we check if the Zip utility is available. If not, nothing can be done.
        if (!class_exists('ZipArchive'))
        {
            utilities::fl("ZipArchive not supported for this host. The plugin cannot unzip the file. Process aborted","");
            return;
        }
		//Proceed to download zip data
        utilities::fl("Downloading zip file ", "");
        $ua_file= self::remoteRequest(self::$_zip_url);
        if(is_wp_error($ua_file))
        {
        	//Analyze error type
        	$code   = wp_remote_retrieve_response_code($ua_file);
			$errMsg = wp_remote_retrieve_response_message($ua_file);
            utilities::fl("Code returned:" , $code);
        	if ($code <500) //Request failed of a client error
        	{
        		utilities::fl("Client error $code while getting zip file: " , $errMsg);
				return;
        	}
        	else //Request failed because a server error
        	{
				utilities::fl("Server error $code: while getting zip file: ", $errMsg);
				return;
        	}
        }
        utilities::fl("File downloaded proceed saving to disk","");
		//File downloaded, save to disk. TODO: discuss if we have to use @
		@file_put_contents($PARSER_ZIP, wp_remote_retrieve_body($ua_file));
		if (!file_exists($PARSER_ZIP))
		{
			//TODO:Check permision problem to alert user properly
			utilities::fl("File $PARSER_ZIP not saved: permission problem?","");
			return;
		}
        utilities::fl("File saved ok, proceed to check with md5");
		//Zip file in disk, check md5 with the web service (they have to be syncrhonized)
		$md5response=self::remoteRequest (self::$_md5_url);
        //check if error
        if(is_wp_error($md5response))
        {
        	//Analyze error type
        	$code= wp_remote_retrieve_response_code($md5response);
        	if ($code <500) //Request failed of a client error
        	{
        		utilities::fl("Client error $code while checking md5 on server: " . wp_remote_retrieve_response_message($md5response));
				return;
        	}
        	else //Request failed because a server error
        	{
				utilities::fl("Server error $code while getting md5 on server: " . wp_remote_retrieve_response_message($md5response));
				return;
        	}
        }
        utilities::fl("response md5response:",$md5response);
		$md5_from_server= wp_remote_retrieve_body($md5response);
		$md5_file=md5_file($PARSER_ZIP);
		if ($md5_from_server!=$md5_file)
		{
			utilities::fl("Check file mismatch: $md5_from_server != $md5_file ");
			return;
		}
        utilities::fl("File checked ok, proceed to unzip","");
		//Proceed to unzip the file
		$zip = new ZipArchive() ;
		//open archive
		//TODO: check if this is really a zip file
		if ($zip->open($PARSER_ZIP) !== TRUE)
		{
			utilities::fl('Could not open zip file. Aborting...');
			return;
		}
        utilities::fl("File zip downloaded ok, proceed to unzip");
		//Extract contents to destination directory
		//Bug in windows: files using \ confuse the class, forcing to make a folder
		$realpath=str_replace("\\","/",dirname($PARSER_FILE));  //watchout:there is no need to put filename.
		//TODO:the destination filename has to be exactly the name we are expecting. Check or see a way to put exactly the name we're expecting.
		$zip->extractTo($realpath);
		$zip->close();
		//see also http://php.net/manual/es/ref.zip.php
		//and also php zip stream http://php.net/manual/en/wrappers.compression.php
		//Checking if the file is extracted right.
		if (!file_exists($PARSER_FILE))
		{
			utilities::fl("uasdata creation failed: permission problems?");
			return;
		}
        utilities::fl("file unzipped and in correct directory" , $PARSER_FILE);
		//TODO: until now I don't know yet if the file is valid or was tampered (since I only checked the ZIP file not the inside.)
        @unlink($PARSER_ZIP);
        if (file_exists($PARSER_ZIP))
        {
            utilities::fl("File $PARSER_ZIP not deleted: permission problem?","");
            //return; not critical error at this stage
        }
        else
        {
            utilities::fl("File $PARSER_ZIP deleted ok");
        }

        //Get version from server
        $version= self::remoteRequest(self::$_ver_url);
        //check if error
        if ($this->versionError($version))
            return;

        $ver=wp_remote_retrieve_body($version);
        //Make cache file with the new version
        $this->buildCacheFile($PARSER_CACHE,$ver); //true if success, false if not.
	}


    /**
     * Make the request to the restful service
     * @param $url
     */
    static function remoteRequest($url)
    {
        global $wp_version;
        $response= wp_remote_request($url,array('timeout'=>self::$TIMEOUT,
            'user-agent' =>  "WordPress/$wp_version;" . get_bloginfo('url') . ";" . self::$plugin_version));
        return $response;
    }
    /**
     * return true if there is an error when retrieving the version
     * @param $version_server
     * @return bool
     */
    private function versionError($version_server)
    {
        if(is_wp_error($version_server))
        {
            //Analyze error type
            $code   = wp_remote_retrieve_response_code($version_server);
            $errMsg = wp_remote_retrieve_response_message($version_server);
            utilities::fl("error code:" , $code);
            utilities::fl("error detail:",$version_server);
            if ($code <500) //Request failed of a client error
            {
                utilities::fl("Client error $code while getting version: " , $errMsg);
                return true;
            }
            else //Request failed because a server error
            {
                utilities::fl("Server error $code: while getting version: ", $errMsg);
                return true;
            }
        }
        return false;
    }

    /**
     * Build Cache file, return false if fails, true if succeed
     * @param $PARSER_CACHE: Location of where the cache will be stored
     * @param $ver: File version
     * @return void
     */
    private function buildCacheFile($PARSER_CACHE,$ver)
    {
        // Get date from WP
        $timestamp = current_time('timestamp');
        $vdate = gmdate("Ymd", $timestamp);
        $vtime = gmdate("H:i:s", $timestamp);

        // Build a new cache file and store it
        $cacheIni = "; Cache info for class UASparser. Built time: $vdate $vtime";
        $cacheIni .= "[main]\n";
        $cacheIni .= "localversion = \"$ver\"\n";
        $cacheIni .= 'lastupdate = "'.time()."\"\n";
        @file_put_contents($PARSER_CACHE, $cacheIni);

        //Check if cache was built correctly
        if (!file_exists($PARSER_CACHE))
        {
            utilities::fl("Cache file creation failed or not found: permission problems?");
            return; //not necessary but if we expand the method it would has sense.
        }
        utilities::fl("file $PARSER_CACHE saved","");
        //assign current_time from cmb_uarefresh
        $options=get_option("statcomm_options");
        //default two week if cmb undefined.
        //reset current_time
        $options['current_lapse'] = isset($options['cmb_uarefresh'])?$options['cmb_uarefresh']:1209600;
        update_option("statcomm_options",$options);
        return;
    }

    /**
     *  Set the cache directory
     *  @param string
     */
    public function SetCacheDir($cache_dir) {

        // The directory does not exist at this moment, try to make it
        //Not anymore
        // if (!file_exists($cache_dir)) @mkdir($cache_dir, 0777, true);

        // perform some extra checks
        if (!is_writable($cache_dir) || !is_dir($cache_dir)){
            trigger_error('ERROR: Cache dir('.$cache_dir.') is not a directory or not writable');
            return;
        }

        // store the cache dir
        $cache_dir = realpath($cache_dir);
        $this->_cache_dir = $cache_dir;
    }
}
?>